// ===========================================================================
// CONTENT  : CLASS DelegatingInvocationHandler
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 21/07/2013
// HISTORY  :
//  10/10/2008  mdu  CREATED
//	21/07/2013	mdu		changed	--> generic type for delegate
//
// Copyright (c) 2008-2013, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.extender;

//===========================================================================
//IMPORTS
//===========================================================================
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

/**
 * A generic InvocationHandler that passes all method calls through to a
 * delegate object. If that delegate object is not available initially then
 * an ObjectRetrievalService and an identifier will be used to get it.
 *
 * @author M.Duchrow
 * @version 1.1
*/
public class DelegatingInvocationHandler<TDelegate> implements InvocationHandler
{
	// =========================================================================
	// INSTANCE VARIABLES
	// =========================================================================
	private TDelegate delegate = null;

	// =========================================================================
	// CONSTRUCTORS
	// =========================================================================
	protected DelegatingInvocationHandler()
	{
		super();
	} // DelegatingInvocationHandler() 

	// -------------------------------------------------------------------------

	/**
	 * Initialize the new instance with an object to delegate all method calls to.
	 * 
	 * @param object The object to which method calls are delegated.
	 * @throws IllegalArgumentException if object is null.
	 */
	public DelegatingInvocationHandler(TDelegate object)
	{
		this();
		if (object == null)
		{
			throw new IllegalArgumentException("Delegation object must not be null!");
		}
		this.setDelegate(object);
	} // DelegatingInvocationHandler() 

	// -------------------------------------------------------------------------

	// =========================================================================
	// PUBLIC INSTANCE METHODS
	// =========================================================================
	/**
	 * Executes the given method with the specified args against the internal
	 * delegate.
	 */
	@Override
  public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
	{
		return method.invoke(this.getDelegate(), args);
	} // invoke() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the object to which the requested method calls will be delegated.
	 */
	public TDelegate getDelegate()
	{
		return delegate;
	} // getDelegate()

	// -------------------------------------------------------------------------

	// =========================================================================
	// PROTECTED INSTANCE METHODS
	// =========================================================================
	protected void setDelegate(TDelegate delegate)
	{
		this.delegate = delegate;
	} // setDelegate()

	// -------------------------------------------------------------------------

} // class DelegatingInvocationHandler 
