// ===========================================================================
// CONTENT  : CLASS ExtendedObject
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 21/07/2013
// HISTORY  :
//  22/11/2008  mdu  CREATED
//	14/11/2009	mdu		added		-->	isSameMethod()
//	27/12/2009	mdu		changed	-->	ensured that delegate can override any method
//	21/07/2013	mdu		changed	--> generic type for delegate
//
// Copyright (c) 2008-2013, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.extender;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.pfsw.reflect.ReflectUtil;

/**
 * This is the core of the extendible objects framework.
 * This handler is responsible for holding extensions and mix-ins and to dispatch
 * method calls to the correct receiver.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class ExtendedObject<TDelegate> extends DelegatingInvocationHandler<TDelegate>
{
	// =========================================================================
	// CONSTANTS
	// =========================================================================

	// =========================================================================
	// INSTANCE VARIABLES
	// =========================================================================
	private Map<String, Object> extensions = new HashMap<String, Object>();
	protected Map<String, Object> getExtensions()
	{
		return extensions;
	}
	protected void setExtensions(Map<String, Object> newValue)
	{
		extensions = newValue;
	}

	private Map<Class, Object> mixins = new HashMap<Class, Object>();
	protected Map<Class, Object> getMixins()
	{
		return mixins;
	}
	protected void setMixins(Map<Class, Object> newValue)
	{
		mixins = newValue;
	}

	// =========================================================================
	// CONSTRUCTORS
	// =========================================================================
	public ExtendedObject(TDelegate object)
	{
		super(object);
	} // ExtendedObject() 

	// =========================================================================
	// PUBLIC INSTANCE METHODS
	// =========================================================================
	@Override
	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
	{
		Object target;
		target = this.findTargetOf(method);
		if (target == null)
		{
			throw new NoSuchMethodException("Method " + method + " not supported.");
		}
		return method.invoke(target, args);
	} // invoke() 

	// -------------------------------------------------------------------------

	/**
	 * Attach the given extension to the object this invocation handler stands for.
	 * 
	 * @param name The name of the extension (must not be null)
	 * @param extension Any object that must be attached (must not be null)
	 */
	public void setExtension(String name, Object extension)
	{
		if ((name != null) && (extension != null))
		{
			this.getExtensions().put(name, extension);
		}
	} // setExtension() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the extension object for the given name or null if not available.
	 */
	public Object getExtension(String name)
	{
		if (name == null)
		{
			return null;
		}
		return this.getExtensions().get(name);
	} // getExtension() 

	// -------------------------------------------------------------------------

	/**
	 * Remove and return the extension object for the given name or 
	 * return null if not available.
	 */
	public Object removeExtension(String name)
	{
		if (name == null)
		{
			return null;
		}
		return this.getExtensions().remove(name);
	} // removeExtension() 

	// -------------------------------------------------------------------------

	/**
	 * Sets the implementation instance for the specified interface.
	 * If there is another implementation set for this interface already, it will
	 * be removed.
	 * <p>
	 * The given proxy object will be set to all fields of the implementation object
	 * that are annotated by @Self.
	 * <p>
	 * If any of the given parameters is null, nothing will be set at all.
	 */
	public <I, T extends I> void setMixin(Proxy proxy, Class<I> interfaceClass, T implementation)
	{
		if ((proxy != null) && (interfaceClass != null) && (implementation != null))
		{
			this.getMixins().put(interfaceClass, implementation);
			this.injectBackReferenceIntoMixin(proxy, implementation);
		}
	} // setMixin() 

	// -------------------------------------------------------------------------

	public <I> I getMixin(Class<I> interfaceClass)
	{
		if (interfaceClass.isInstance(this.getDelegate()))
		{
			return null;
		}
		return (I)this.getMixins().get(interfaceClass);
	} // getMixin() 

	// -------------------------------------------------------------------------

	/**
	 * Removes the mix-in implementation for the given interface and returns it.
	 * If null is returned, no mix-in for this interface was found.
	 */
	public <I> I removeMixin(Class<I> interfaceClass)
	{
		if ((interfaceClass == null) || (interfaceClass.isInstance(this.getDelegate())))
		{
			return null;
		}
		return (I)this.getMixins().remove(interfaceClass);
	} // removeMixin() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the implementing object of the given interface or null if none
	 * is found.
	 * The result could be the underlying delegate as well as a mix-in.
	 */
	public <I> I getImplementor(Class<I> interfaceClass)
	{
		if (interfaceClass.isInstance(this.getDelegate()))
		{
			return (I)this.getDelegate();
		}
		return (I)this.getMixins().get(interfaceClass);
	} // getImplementor() 

	// -------------------------------------------------------------------------

	// =========================================================================
	// PROTECTED INSTANCE METHODS
	// =========================================================================
	protected Object findTargetOf(Method method)
	{
		if (this.implementsMethod(this.getDelegate().getClass(), method))
		{
			return this.getDelegate();
		}
		if (this.understandsMethod(this.getDelegate(), method))
		{
			return this.getDelegate();
		}
		for (Map.Entry<Class, Object> entry : this.getMixins().entrySet())
		{
			if (this.understandsMethod(entry.getKey(), method))
			{
				return entry.getValue();
			}
		}
		return this.getDelegate();
	} // findTargetOf() 

	// -------------------------------------------------------------------------

	protected boolean understandsMethod(Object object, Method method)
	{
		Class[] interfaces;

		if (object == null)
		{
			return false;
		}
		interfaces = object.getClass().getInterfaces();
		for (int i = 0; i < interfaces.length; i++)
		{
			if (this.understandsMethod(interfaces[i], method))
			{
				return true;
			}
		}
		return false;
	} // understandsMethod() 

	// -------------------------------------------------------------------------

	/**
	 * Returns true if the given interface or class is understanding the given
	 * method, either by declaring it itself or by inheritance.
	 */
	protected boolean understandsMethod(Class interfaceOrClass, Method method)
	{
		Method[] methods;

		methods = interfaceOrClass.getMethods();
		return containsMethod(methods, method);
	} // understandsMethod() 

	// -------------------------------------------------------------------------

	/**
	 * Returns true if the given class is no interface and is implementing the given
	 * method itself, that is not inheriting it.
	 */
	protected boolean implementsMethod(Class interfaceOrClass, Method method)
	{
		Method[] methods;

		if (interfaceOrClass.isInterface())
		{
			return false;
		}
		methods = interfaceOrClass.getDeclaredMethods();
		return this.containsMethod(methods, method);
	} // implementsMethod() 

	// -------------------------------------------------------------------------

	/**
	 * Returns true if the given method array contains the specified method. 
	 */
	protected boolean containsMethod(Method[] methods, Method method)
	{
		boolean isMethodOfObject;

		isMethodOfObject = (Object.class == method.getDeclaringClass());
		for (int i = 0; i < methods.length; i++)
		{
			if (isMethodOfObject) // This allows to overwrite methods declared in java.lang.Object
			{
				if (this.isSameMethod(method, methods[i]))
				{
					return true;
				}
			}
			else if (method.equals(methods[i]))
			{
				return true;
			}
		}
		return false;
	} // containsMethod() 

	// -------------------------------------------------------------------------

	/**
	 * Returns true if the two given methods are the same except that they do 
	 * not have to be declared by the same class or interface.
	 */
	protected boolean isSameMethod(Method method1, Method method2)
	{
		if ((method1.getName().equals(method2.getName())))
		{
			if (!method1.getReturnType().equals(method2.getReturnType()))
			{
				return false;
			}
			/* Avoid unnecessary cloning */
			Class[] params1 = method1.getParameterTypes();
			Class[] params2 = method2.getParameterTypes();
			if (params1.length == params2.length)
			{
				for (int i = 0; i < params1.length; i++)
				{
					if (params1[i] != params2[i])
						return false;
				}
				return true;
			}
		}
		return false;
	} // isSameMethod() 

	// -------------------------------------------------------------------------

	private void injectBackReferenceIntoMixin(Proxy proxy, Object mixin)
	{
		List<Field> fields;
		Class[] interfaces;

		interfaces = ReflectUtil.current().getInterfacesOf(proxy);
		fields = ReflectUtil.current().getFieldsOf(mixin); // Includes all inherited fields!
		for (Field field : fields)
		{
			if (field.isAnnotationPresent(Self.class))
			{
				if (ReflectUtil.current().contains(interfaces, field.getType()))
				{
					try
					{
						ReflectUtil.current().setValueOf(mixin, field.getName(), proxy);
					}
					catch (NoSuchFieldException ex)
					{
						// Should not happen, because the field was just detected in the object.
						ex.printStackTrace();
					}
				}
			}
		}
	} // injectBackReferenceIntoMixin() 

	// -------------------------------------------------------------------------

} // class ExtendedObject 
