// ===========================================================================
// CONTENT  : CLASS ObjectExtender
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 14/11/2009
// HISTORY  :
//  22/11/2008  mdu  CREATED
//	14/11/2009	mdu		changed	-->	Allowed additional interfaces at proxy generation
//
// Copyright (c) 2008-2009, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.extender;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;

/**
 * This class provides various static methods for creation of "extended" objects
 * and to add mix-ins.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public class ObjectExtender
{
	// =========================================================================
	// CONSTANTS
	// =========================================================================

	// =========================================================================
	// CLASS METHODS
	// =========================================================================
	/**
	 * Returns the given object enhanced by a change tracking mechanism.
	 * The actual return type is the specified interface type.
	 * 
	 * @param interfaceType The interface the provided object must implement
	 * @param object An object that should be enhanced by a change tracking mechanism
	 */
	public static <I> I makeExtensible(Class<I> interfaceType, I object, Class... interfaces)
	{
		I result;
		ExtendedObject extensionHandler;
		Class[] allInterfaces;

		extensionHandler = ObjectExtender.getExtensionHandlerOf(object);
		if (extensionHandler == null)
		{
			allInterfaces = toOneArray(interfaceType, interfaces);
			extensionHandler = new ExtendedObject(object);
			result = (I)Proxy.newProxyInstance(object.getClass().getClassLoader(), allInterfaces, extensionHandler);
		}
		else
		{
			result = object;
		}
		return result;
	} // makeExtensible() 

	// -------------------------------------------------------------------------

	/**
	 * Returns an extended object that implements the given interface and is based on
	 * the given object.
	 * 
	 * @param interfaceType The interface the extended object implements (must not be null)
	 * @param object The object to be extended (must not be null)
	 */
	public static <I> I createExtensibleObject(Class<I> interfaceType, Object object)
	{
		Class[] allInterfaces;

		allInterfaces = new Class[] { interfaceType };
		return (I)createExtensibleObject(object, allInterfaces);
	} // createExtensibleObject() 

	// -------------------------------------------------------------------------

	/**
	 * Returns an extended object that implements the given interface and is based on
	 * the given object. It also supports the additional interfaces, which are however 
	 * not part of the returned type.
	 * 
	 * @param interfaceType The interface the extended object implements (must not be null)
	 * @param object The object to be extended (must not be null)
	 * @param additionalInterfaces More interface that are not part of the return type
	 */
	public static <I> I createExtensibleObject(Class<I> interfaceType, Object object, Class... additionalInterfaces)
	{
		Class[] allInterfaces;

		allInterfaces = toOneArray(interfaceType, additionalInterfaces);
		return (I)createExtensibleObject(object, allInterfaces);
	} // createExtensibleObject() 

	// -------------------------------------------------------------------------

	/**
	 * Returns an object of the specified interface type that also implements
	 * the additional interfaces.
	 * The implementation instances for all interfaces must be added as  mix-ins.
	 * 
	 * @param interfaceType The interface the provided object must implement
	 * @param interfaces Optional additional interfaces
	 */
	public static <I> I createExtensibleObject(Class<I> interfaceType, Class... interfaces)
	{
		I result;
		Class[] allInterfaces;

		allInterfaces = toOneArray(interfaceType, interfaces);
		result = (I)createExtensibleObject(new Object(), allInterfaces);
		return result;
	} // createExtensibleObject() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the mix-in implementation that is associated with the given
	 * interface and attached to the provided object.
	 * If no such mix-in can be found or the given object is not an extended
	 * object then this method returns null.
	 * 
	 * @param interfaceClass The interface of the mix-in
	 * @param object The extended object that might contain the mix-in
	 */
	public static <T> T getMixin(Class<T> interfaceClass, Object object)
	{
		ExtendedObject extensionHandler;

		extensionHandler = getExtensionHandlerOf(object);
		if (extensionHandler != null)
		{
			return (T)extensionHandler.getMixin(interfaceClass);
		}
		return null;
	} // getMixin() 

	// -------------------------------------------------------------------------

	/**
	 * Sets the mix-in implementation of the specified interface to the provided
	 * object. If any of the given parameters is null, nothing happens.
	 */
	public static <I, T extends I> void setMixinTo(Object object, Class<I> interfaceType, T mixin)
	{
		ExtendedObject extendedObject;

		if (isNull(object, interfaceType, mixin))
		{
			return;
		}
		extendedObject = getExtensionHandlerOf(object);
		if (extendedObject != null)
		{
			extendedObject.setMixin((Proxy)object, interfaceType, mixin);
		}
	} // setMixinTo() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the ExtendedObject that is attached to the
	 * given object or null if none is attached.
	 */
	public static ExtendedObject getExtensionHandlerOf(Object object)
	{
		Proxy proxy;
		InvocationHandler invocationHandler;

		if (object instanceof Proxy)
		{
			proxy = (Proxy)object;
			invocationHandler = Proxy.getInvocationHandler(proxy);
			if (invocationHandler instanceof ExtendedObject)
			{
				return (ExtendedObject)invocationHandler;
			}
		}
		return null;
	} // getExtensionHandlerOf() 

	// -------------------------------------------------------------------------

	/**
	 * Returns the named extension associated with the given object or
	 * null if there is none.
	 */
	public static Object getExtensionOf(Object object, String extensionName)
	{
		ExtendedObject extensionHandler;

		extensionHandler = ObjectExtender.getExtensionHandlerOf(object);
		if (extensionHandler == null)
		{
			return null;
		}
		return extensionHandler.getExtension(extensionName);
	} // getExtensionOf() 

	// -------------------------------------------------------------------------

	/**
	 * Returns an extended object that implements all of the given interfaces and is 
	 * based on the given object. 
	 * 
	 * @param object The object to be extended (must not be null)
	 * @param allInterfaces All interfaces the returned extended object will support
	 */
	private static Object createExtensibleObject(Object object, Class[] allInterfaces)
	{
		ExtendedObject extensionHandler;

		extensionHandler = new ExtendedObject(object);
		return Proxy.newProxyInstance(allInterfaces[0].getClassLoader(), allInterfaces, extensionHandler);
	} // createExtensibleObject() 

	// -------------------------------------------------------------------------

	private static Class[] toOneArray(Class interfaceType, Class... interfaces)
	{
		Class[] allInterfaces;

		if (isNullOrEmpty(interfaces))
		{
			allInterfaces = new Class[] { interfaceType };
		}
		else
		{
			allInterfaces = new Class[interfaces.length + 1];
			allInterfaces[0] = interfaceType;
			for (int i = 0; i < interfaces.length; i++)
			{
				allInterfaces[i + 1] = interfaces[i];
			}
		}
		return allInterfaces;
	} // toOneArray() 

	// -------------------------------------------------------------------------

	private static boolean isNullOrEmpty(Class[] interfaces)
	{
		return (interfaces == null) || (interfaces.length == 0);
	} // isNullOrEmpty() 

	// -------------------------------------------------------------------------

	private static boolean isNull(Object obj1, Object obj2, Object obj3)
	{
		return (obj1 == null) || (obj2 == null) || (obj3 == null);
	} // isNull() 

	// -------------------------------------------------------------------------

	// =========================================================================
	// CONSTRUCTORS
	// =========================================================================
	/**
	 * NO INSTANCES SUPPORTED
	 */
	private ObjectExtender()
	{
		super();
	} // ObjectExtender() 

} // class ObjectExtender 
